@@ -73,7 +73,7 @@ var incompleteSentinel = &Bottom{
 // error.
 //
 // TODO: return *Vertex
-func (c *OpContext) evaluate(v *Vertex, r Resolver, state vertexStatus) Value {
+func (c *OpContext) evaluate(v *Vertex, r Resolver, state combinedFlags) Value {
 	if v.isUndefined() {
 		// Use node itself to allow for cycle detection.
 		c.unify(v, state)
@@ -150,7 +150,13 @@ func (c *OpContext) evaluate(v *Vertex, r Resolver, state vertexStatus) Value {
 // state can be used to indicate to which extent processing should continue.
 // state == finalized means it is evaluated to completion. See vertexStatus
 // for more details.
-func (c *OpContext) unify(v *Vertex, state vertexStatus) {
+func (c *OpContext) unify(v *Vertex, flags combinedFlags) {
+	if c.isDevVersion() {
+		requires, mode := flags.conditions(), flags.runMode()
+		v.unify(c, requires, mode)
+		return
+	}
+
 	// defer c.PopVertex(c.PushVertex(v))
 	if Debug {
 		c.nest++
@@ -166,6 +172,8 @@ func (c *OpContext) unify(v *Vertex, state vertexStatus) {
 	n := v.getNodeContext(c, 1)
 	defer v.freeNode(n)
 
+	state := flags.vertexStatus()
+
 	// TODO(cycle): verify this happens in all cases when we need it.
 	if n != nil && v.Parent != nil && v.Parent.state != nil {
 		n.depth = v.Parent.state.depth + 1
@@ -370,6 +378,8 @@ func (c *OpContext) unify(v *Vertex, state vertexStatus) {
 
 // insertConjuncts inserts conjuncts previously not inserted.
 func (n *nodeContext) insertConjuncts(state vertexStatus) bool {
+	unreachableForDev(n.ctx)
+
 	// Exit early if we have a concrete value and only need partial results.
 	if state == partial {
 		for n.conjunctsPartialPos < len(n.conjuncts) {
@@ -461,13 +471,14 @@ func (n *nodeContext) doNotify() {
 
 func (n *nodeContext) postDisjunct(state vertexStatus) {
 	ctx := n.ctx
+	unreachableForDev(ctx)
 
 	for {
 		// Use maybeSetCache for cycle breaking
 		for n.maybeSetCache(); n.expandOne(state); n.maybeSetCache() {
 		}
 
-		if !n.addLists(state) {
+		if !n.addLists(oldOnly(state)) {
 			break
 		}
 	}
@@ -735,6 +746,8 @@ func (n *nodeContext) incompleteErrors(final bool) *Bottom {
 // though, that any potential performance issues are eliminated for
 // Protobuf-like oneOf fields.
 func (n *nodeContext) checkClosed(state vertexStatus) bool {
+	unreachableForDev(n.ctx)
+
 	ignore := state != finalized || n.skipNonMonotonicChecks()
 
 	v := n.node
@@ -754,6 +767,8 @@ func (n *nodeContext) checkClosed(state vertexStatus) bool {
 }
 
 func (n *nodeContext) completeArcs(state vertexStatus) {
+	unreachableForDev(n.ctx)
+
 	if DebugSort > 0 {
 		DebugSortArcs(n.ctx, n.node)
 	}
@@ -790,7 +805,7 @@ func (n *nodeContext) completeArcs(state vertexStatus) {
 
 			wasVoid := a.ArcType == ArcPending
 
-			ctx.unify(a, finalized)
+			ctx.unify(a, oldOnly(finalized))
 
 			if a.ArcType == ArcPending {
 				continue
@@ -844,7 +859,7 @@ func (n *nodeContext) completeArcs(state vertexStatus) {
 			// TODO(errors): make Validate return bottom and generate
 			// optimized conflict message. Also track and inject IDs
 			// to determine origin location.s
-			v := ctx.evalState(c.expr, finalized)
+			v := ctx.evalState(c.expr, oldOnly(finalized))
 			v, _ = ctx.getDefault(v)
 			v = Unwrap(v)
 
@@ -952,17 +967,13 @@ type nodeContext struct {
 	nextFree *nodeContext
 	refCount int
 
-	// Keep these two out of the nodeContextState to make them more accessible
+	// Keep node out of the nodeContextState to make them more accessible
 	// for source-level debuggers.
-	ctx  *OpContext
 	node *Vertex
 
 	nodeContextState
 
-	// rootCloseContext should not be cloned as clones need to get their own
-	// copies of this. For this reason it is not included in nodeContextState,
-	// as it prevents it from being set "by default".
-	rootCloseContext *closeContext
+	scheduler
 
 	// Below are slices that need to be managed when cloning and reclaiming
 	// nodeContexts for reuse. We want to ensure that, instead of setting
@@ -1046,6 +1057,11 @@ type nodeContextState struct {
 	aStruct   Expr
 	aStructID CloseInfo
 
+	// List fields
+	listIsClosed bool
+	maxListLen   int
+	maxNode      Expr
+
 	lowerBound *BoundValue // > or >=
 	upperBound *BoundValue // < or <=
 	errs       *Bottom
@@ -1059,6 +1075,8 @@ type nodeContextState struct {
 	// conjunctsPartialPos is like conjunctsPos, but for the 'partial' phase
 	// of processing where conjuncts are only processed as concrete scalars.
 	conjunctsPartialPos int
+
+	arcPos int
 }
 
 // A receiver receives notifications.
@@ -1086,6 +1104,8 @@ type defaultInfo struct {
 }
 
 func (n *nodeContext) addNotify(v *Vertex, cc *closeContext) {
+	unreachableForDev(n.ctx)
+
 	if v != nil && !n.node.hasAllConjuncts {
 		n.notify = append(n.notify, receiver{v, cc})
 	}
@@ -1104,6 +1124,8 @@ func (n *nodeContext) clone() *nodeContext {
 	d.arcMap = append(d.arcMap, n.arcMap...)
 	d.notify = append(d.notify, n.notify...)
 
+	n.scheduler.cloneInto(&d.scheduler)
+
 	d.conjuncts = append(d.conjuncts, n.conjuncts...)
 	d.cyclicConjuncts = append(d.cyclicConjuncts, n.cyclicConjuncts...)
 	d.dynamicFields = append(d.dynamicFields, n.dynamicFields...)
@@ -1129,8 +1151,8 @@ func (c *OpContext) newNodeContext(node *Vertex) *nodeContext {
 		c.freeListNode = n.nextFree
 
 		*n = nodeContext{
-			ctx:  c,
-			node: node,
+			scheduler: scheduler{ctx: c},
+			node:      node,
 			nodeContextState: nodeContextState{
 				kind: TopKind,
 			},
@@ -1152,20 +1174,28 @@ func (c *OpContext) newNodeContext(node *Vertex) *nodeContext {
 			disjuncts:          n.disjuncts[:0],
 			buffer:             n.buffer[:0],
 		}
+		n.scheduler.clear()
+		n.scheduler.node = n
 
 		return n
 	}
 	c.stats.Allocs++
 
-	return &nodeContext{
-		ctx:  c,
+	n := &nodeContext{
+		scheduler: scheduler{
+			ctx: c,
+		},
 		node: node,
 
 		nodeContextState: nodeContextState{kind: TopKind},
 	}
+	n.scheduler.node = n
+	return n
 }
 
 func (v *Vertex) getNodeContext(c *OpContext, ref int) *nodeContext {
+	unreachableForDev(c)
+
 	if v.state == nil {
 		if v.status == finalized {
 			return nil
@@ -1219,6 +1249,7 @@ func (c *OpContext) freeNodeContext(n *nodeContext) {
 	c.freeListNode = n
 	n.node = nil
 	n.refCount = 0
+	n.scheduler.clear()
 }
 
 // TODO(perf): return a dedicated ConflictError that can track original
@@ -1321,6 +1352,8 @@ func (n *nodeContext) updateNodeType(k Kind, v Expr, id CloseInfo) bool {
 }
 
 func (n *nodeContext) done() bool {
+	// TODO(v0.7): verify that done() is checking for the right conditions in
+	// the new evaluator implementation.
 	return len(n.dynamicFields) == 0 &&
 		len(n.comprehensions) == 0 &&
 		len(n.exprs) == 0
@@ -1329,6 +1362,7 @@ func (n *nodeContext) done() bool {
 // finalDone is like done, but allows for cycle errors, which can be ignored
 // as they essentially indicate a = a & _.
 func (n *nodeContext) finalDone() bool {
+	// TODO(v0.7): update for new evaluator?
 	for _, x := range n.exprs {
 		if x.err.Code != CycleError {
 			return false
@@ -1474,6 +1508,8 @@ func (n *nodeContext) addErr(err errors.Error) {
 // into the nodeContext if successful or queue it for later evaluation if it is
 // incomplete or is not value.
 func (n *nodeContext) addExprConjunct(v Conjunct, state vertexStatus) {
+	unreachableForDev(n.ctx)
+
 	env := v.Env
 	id := v.CloseInfo
 
@@ -1525,6 +1561,8 @@ func (n *nodeContext) addExprConjunct(v Conjunct, state vertexStatus) {
 // evalExpr is only called by addExprConjunct. If an error occurs, it records
 // the error in n and returns nil.
 func (n *nodeContext) evalExpr(v Conjunct, state vertexStatus) {
+	unreachableForDev(n.ctx)
+
 	// Require an Environment.
 	ctx := n.ctx
 
@@ -1539,7 +1577,7 @@ func (n *nodeContext) evalExpr(v Conjunct, state vertexStatus) {
 		if state == finalized {
 			state = conjuncts
 		}
-		arc, err := ctx.resolveState(v, x, state)
+		arc, err := ctx.resolveState(v, x, oldOnly(state))
 		if err != nil && (!err.IsIncomplete() || err.Permanent) {
 			n.addBottom(err)
 			break
@@ -1576,7 +1614,7 @@ func (n *nodeContext) evalExpr(v Conjunct, state vertexStatus) {
 	case Evaluator:
 		// Interpolation, UnaryExpr, BinaryExpr, CallExpr
 		// Could be unify?
-		val := ctx.evaluateRec(v, partial)
+		val := ctx.evaluateRec(v, oldOnly(partial))
 		if b, ok := val.(*Bottom); ok &&
 			b.IsIncomplete() {
 			n.exprs = append(n.exprs, envExpr{v, b})
@@ -1616,6 +1654,8 @@ func (n *nodeContext) evalExpr(v Conjunct, state vertexStatus) {
 }
 
 func (n *nodeContext) addVertexConjuncts(c Conjunct, arc *Vertex, inline bool) {
+	unreachableForDev(n.ctx)
+
 	closeInfo := c.CloseInfo
 
 	// We need to ensure that each arc is only unified once (or at least) a
@@ -1689,7 +1729,7 @@ func (n *nodeContext) addVertexConjuncts(c Conjunct, arc *Vertex, inline bool) {
 		// is necessary to prevent lookups in unevaluated structs.
 		// TODO(cycles): this can probably most easily be fixed with a
 		// having a more recursive implementation.
-		n.ctx.unify(arc, partial)
+		n.ctx.unify(arc, oldOnly(partial))
 	}
 
 	// Don't add conjuncts if a node is referring to itself.
@@ -2021,6 +2061,10 @@ func (n *nodeContext) addStruct(
 // disjunctions.
 func (n *nodeContext) insertField(f Feature, mode ArcType, x Conjunct) *Vertex {
 	ctx := n.ctx
+	if ctx.isDevVersion() {
+		return n.insertArc(f, mode, x, x.CloseInfo, true)
+	}
+
 	arc, isNew := n.node.GetArc(ctx, f, mode)
 	if f.IsLet() && !isNew {
 		arc.MultiLet = true
@@ -2050,6 +2094,10 @@ func (n *nodeContext) insertField(f Feature, mode ArcType, x Conjunct) *Vertex {
 
 func (n *nodeContext) insertFieldUnchecked(f Feature, mode ArcType, x Conjunct) *Vertex {
 	ctx := n.ctx
+	if ctx.isDevVersion() {
+		return n.insertArc(f, mode, x, x.CloseInfo, false)
+	}
+
 	arc, isNew := n.node.GetArc(ctx, f, mode)
 	if f.IsLet() && !isNew {
 		arc.MultiLet = true
@@ -2073,6 +2121,8 @@ func (n *nodeContext) insertFieldUnchecked(f Feature, mode ArcType, x Conjunct)
 // TODO(errors): detect when a field is added to a struct that is already used
 // in a for clause.
 func (n *nodeContext) expandOne(state vertexStatus) (done bool) {
+	unreachableForDev(n.ctx)
+
 	// Don't expand incomplete expressions if we detected a cycle.
 	if n.done() || (n.hasCycle && !n.hasNonCycle) {
 		return false
@@ -2110,6 +2160,8 @@ func (n *nodeContext) expandOne(state vertexStatus) (done bool) {
 
 // injectDynamic evaluates and inserts dynamic declarations.
 func (n *nodeContext) injectDynamic() (progress bool) {
+	unreachableForDev(n.ctx)
+
 	ctx := n.ctx
 	k := 0
 
@@ -2119,7 +2171,7 @@ func (n *nodeContext) injectDynamic() (progress bool) {
 		x := d.field.Key
 		// Push state to capture and remove errors.
 		s := ctx.PushState(d.env, x.Source())
-		v := ctx.evalState(x, finalized)
+		v := ctx.evalState(x, oldOnly(finalized))
 		b := ctx.PopState(s)
 
 		if b != nil && b.IsIncomplete() {
@@ -2160,7 +2212,7 @@ func (n *nodeContext) injectDynamic() (progress bool) {
 //
 // TODO(embeddedScalars): for embedded scalars, there should be another pass
 // of evaluation expressions after expanding lists.
-func (n *nodeContext) addLists(state vertexStatus) (progress bool) {
+func (n *nodeContext) addLists(state combinedFlags) (progress bool) {
 	if len(n.lists) == 0 && len(n.vLists) == 0 {
 		return false
 	}
@@ -2246,7 +2298,7 @@ outer:
 				if err != nil {
 					if err.ForCycle && !l.self {
 						// The list has a comprehension that refers to the list
-						// itself. This means we should postpone evalauting this
+						// itself. This means we should postpone evaluating this
 						// list until all other lists have been evaluated.
 						n.lists[i].ignore = true
 						l.self = true
